/*
Test program for BER comparison between hard and soft decision
algorithms for BCH 32.21
(c)2000 VLV
vlv@writeme.com
*/



#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "bch_m.h"

#define   Noise         0.55            // Noise dispersion


unsigned int rndseed[4] = { 0x12345678,0x87654321,0xAA55,0x1E };


/*
BCH 32.21 correction. Algorithm (c) K. Borisov
*/

//**************** Aux ***********************************

void InvertBit(unsigned int *bchword,unsigned char bitno)
{
*bchword^=1lu<<bitno;
}

//**************** Aux ************************************

unsigned char MakeSyndrome(unsigned int x, unsigned char polynom)
{
unsigned char cntr;
unsigned long int l_polynom;

x&=0xfffffffelu; // clear parity bit

l_polynom=((unsigned long int)polynom)<<27; // left align the polynom

// Calculate the syndrome

for(cntr=0;cntr<31;cntr++)
 {
 if(x&0x80000000lu) x=(x<<1)^l_polynom;
 else x<<=1;
 }

return (unsigned char)(x>>27);  // right align the syndrome
}

//******************* Correction itself *******************

unsigned char CorrectBCH(unsigned int *bchword)
{
unsigned char sa,sb,parity,dist;
unsigned long int ltmp;

// Syndrome A error position table

static unsigned char table_sa[32] =
{
0x00,0x1b,0x1c,0x0e,0x1d,0x01,0x0f,0x07,
0x1e,0x19,0x02,0x17,0x10,0x04,0x08,0x13,
0x1f,0x06,0x1a,0x0d,0x03,0x12,0x18,0x16,
0x11,0x15,0x05,0x0c,0x09,0x0a,0x14,0x0b
};

// Syndrome B error position table

static unsigned char table_sb[32] =
{
0x00,0x1b,0x1c,0x10,0x1d,0x05,0x11,0x02,
0x1e,0x16,0x06,0x0c,0x12,0x14,0x03,0x19,
0x1f,0x0e,0x17,0x0a,0x07,0x08,0x0d,0x09,
0x13,0x18,0x15,0x0b,0x04,0x01,0x1a,0x0f
};

// Distance between errors table

static unsigned char table_dist[32] =
{
0xff,0xff,0xff,0x03,0xff,0xff,0x06,0x0b,
0xff,0xff,0xff,0xff,0x0c,0xff,0x09,0x08,
0xff,0x0e,0xff,0x0a,0xff,0xff,0xff,0x04,
0x07,0x05,0xff,0x02,0x0d,0x01,0x0f,0xff
};

// Correction table [distance]

static unsigned char dist_correction_table[16] =
{
0xff,0x13,0x06,0x1e,0x0b,0x03,0x1c,0x17,
0x15,0x11,0x05,0x14,0x18,0x0f,0x0e,0x19
};


//========= From Here ===========

ltmp=*bchword;

//  Calculate syndromes A and B

sa=MakeSyndrome(ltmp,0x05);
sb=MakeSyndrome(ltmp,0x1d);

// Calculate parity

parity=0;
while(ltmp)
 {
 parity^=ltmp&1;
 ltmp>>=1;
 }

// Check syndromes

if((sa==0)&&(sb!=0)) return 0xff;
if((sb==0)&&(sa!=0)) return 0xff;


if((sa==0)&&(sb==0)) // If both syndromes are 0 (no code errors)
 {
 if(parity==0) return 0; // No errors
 *bchword^=1;            // Error in parity bit
 return 1;
 }


sa=table_sa[sa];
sb=table_sb[sb];


if(sa==sb) // If both syndromes are indicating one error
 {
 InvertBit(bchword,sa);
 if(parity!=0) return 1;   // One error
 *bchword^=1;              // Error in parity bit
 return 2;
 }

if(parity!=0) return 0xff; // More then two errors - not correctable

dist=sa-sb;

if(dist&0x80) dist+=0x1f; // modulo 1f

dist=table_dist[dist];    // Find distance between errors

if(dist==0xff) return 0xff;      //   Not correctable

sa-=dist_correction_table[dist]; // Correct 1-st error location

if(sa&0x80) sa+=0x1f;            // modulo 31

InvertBit(bchword,sa+1);         // Correct 1-st error

sa+=dist;                       // Find 2-nd error location

if(sa>=31) sa-=31;              //  modulo 31

InvertBit(bchword,sa+1);        // Correct 2-nd error

return 2;
}

/*
Generate 32-bit random value
*/

unsigned int Rnd32(void)
{
unsigned int i,s=0;

for(i=32;i!=0;i--)
 {
 if(rndseed[0]&0x80000000) rndseed[0]=(rndseed[0]<<1)^0xF5;
 else rndseed[0]<<=1;

 rndseed[1]<<=1;
 if(rndseed[1]&0x80000) rndseed[1]^=0x7CF13;

 rndseed[2]<<=1;
 if(rndseed[2]&0x2000) rndseed[2]^=0x0e07;

 rndseed[3]<<=1;
 if(rndseed[3]&0x80) rndseed[3]^=0x77;

 s=(s<<1)|(1&(rndseed[0]^rndseed[1]^rndseed[2]^rndseed[3]));
 }

return s;
}

/*
Simple, Stupid, Slow
Generate double accuracy random number in the 0...1 interval
(Needed for good Gaussian distribution)
*/

double Rnd(void)
{
double s;

s=(((double)Rnd32())*4294967296.0 + ((double)Rnd32()))*5.4210108624275221703311375920e-20;

return s;
}

/*
Generate Gaussian distributed value (dispersion = 1, avg = 0)
*/

float GaussianRnd(void)
{
double r;

r=Rnd();
if(r==0.0) r=2.71e-20;

return sqrt(2.0*log(1.0/r))*cos(6.283185307179586476925286766559*Rnd());
}


//=========================================

void main(void)
{
unsigned int uncoded;

unsigned int i,utmp;
float x[32];

unsigned int nn,N=100000;
unsigned int Ne=0,Nhard=0,Nsum=0;


for(nn=0;nn<N;nn++)
 {
 uncoded =(rand()^(rand()<<8)^(rand()<<16))&0x1FFFFF;
 utmp = EncodeBCH(uncoded);

 for(i=0;i<32;i++)
  {
  x[i] = (utmp&0x80000000) ? 1.0 : -1.0;
  utmp<<=1;
  }

 for(i=0;i<32;i++) x[i]+=Noise*GaussianRnd();

 Nsum+=21;

 utmp = DecodeBCH(x);    // Soft decision
 utmp^=uncoded;

 while(utmp)
  {
  Ne+=utmp&1;
  utmp>>=1;
  }

 utmp = 0;
 for(i=0;i<32;i++)
  {
  utmp<<=1;
  if(x[i]>0.0) utmp|=1;
  }
 CorrectBCH(&utmp);   // Hard decision


 utmp>>=11;

 utmp^=uncoded;

 while(utmp)
  {
  Nhard+=utmp&1;
  utmp>>=1;
  }

 if(!(nn&0x3FF))
  {
  printf("\n Soft BER=%e Hard BER=%e",(float)Ne/(float)Nsum,(float)Nhard/(float)Nsum);
  }
 }
}
